"""Tests for aiotnse helpers module."""
from __future__ import annotations

from base64 import b64encode

from aiohttp import hdrs

from aiotnse.const import (
    DEFAULT_API_HASH,
    DEFAULT_CONTENT_TYPE,
    DEFAULT_USER_AGENT,
    DEVICE_ID_HEADER,
)
from aiotnse.helpers import (
    build_request_headers,
    get_base_url,
    is_error_response,
    is_valid_account,
)


class TestIsValidAccount:
    def test_valid_12_digits(self) -> None:
        assert is_valid_account("610000000001") is True

    def test_too_short(self) -> None:
        assert is_valid_account("000000000") is False

    def test_too_long(self) -> None:
        assert is_valid_account("6100000000011") is False

    def test_empty(self) -> None:
        assert is_valid_account("") is False

    def test_non_digit_chars(self) -> None:
        assert is_valid_account("61000000000a") is False

    def test_with_spaces(self) -> None:
        assert is_valid_account("61000000000 ") is False


class TestBuildRequestHeaders:
    def test_all_headers(self) -> None:
        headers = build_request_headers("rostov", "AP3A.240905.015")
        expected_basic = "Basic " + b64encode(
            b"mobile-api-rostov:mobile-api-rostov"
        ).decode()

        assert headers[hdrs.USER_AGENT] == DEFAULT_USER_AGENT
        assert headers[hdrs.CONTENT_TYPE] == DEFAULT_CONTENT_TYPE
        assert headers["x-api-hash"] == DEFAULT_API_HASH
        assert headers[hdrs.AUTHORIZATION] == expected_basic
        assert headers[DEVICE_ID_HEADER] == "AP3A.240905.015"


class TestGetBaseUrl:
    def test_various_regions(self) -> None:
        assert get_base_url("rostov") == "https://mobile-api-rostov.tns-e.ru"
        assert get_base_url("penza") == "https://mobile-api-penza.tns-e.ru"
        assert get_base_url("nn") == "https://mobile-api-nn.tns-e.ru"


class TestIsErrorResponse:
    def test_success(self) -> None:
        response = {"result": True, "statusCode": 200, "data": []}
        assert is_error_response(response) is False

    def test_false_result(self) -> None:
        response = {"result": False, "statusCode": 400, "error": {"description": "Error"}}
        assert is_error_response(response) is True

    def test_bad_status_code(self) -> None:
        response = {"result": True, "statusCode": 500, "data": []}
        assert is_error_response(response) is True

    def test_missing_result(self) -> None:
        assert is_error_response({}) is True

    def test_non_dict_types(self) -> None:
        assert is_error_response("string") is True
        assert is_error_response(None) is True
        assert is_error_response([]) is True



"""TNS-Energo API wrapper."""
from __future__ import annotations

from typing import Any

from aiohttp import ClientResponseError, ClientSession

from .auth import AbstractTNSEAuth
from .const import (
    DEFAULT_API_PATH,
    DEFAULT_APP_VERSION,
    DEFAULT_PLATFORM,
    DEFAULT_REGION,
    DEVICE_ID,
)
from .exceptions import RequiredApiParamNotFound, TNSEApiError
from .helpers import build_request_headers, get_base_url


async def async_get_regions(session: ClientSession) -> dict[str, Any]:
    """Get available regions.

    Standalone function that does not require authentication.
    Uses the default region endpoint as a bootstrap host.
    """
    base_url = get_base_url(DEFAULT_REGION)
    url = f"{base_url}/{DEFAULT_API_PATH}/contacts/regions"
    headers = build_request_headers(DEFAULT_REGION, DEVICE_ID)
    try:
        async with session.get(url, headers=headers, raise_for_status=True) as resp:
            return await resp.json()
    except ClientResponseError as err:
        raise TNSEApiError(
            f"API request failed: {err.status} {err.message}"
        ) from err


class TNSEApi:
    """TNS-Energo API client."""

    def __init__(self, auth: AbstractTNSEAuth) -> None:
        self._auth = auth

    async def _async_get(
        self, path: str, params: dict[str, Any] | None = None
    ) -> dict[str, Any]:
        """Make GET request to API endpoint."""
        return await self._auth.request("GET", path, params=params)

    async def _async_post(
        self, path: str, json_data: dict[str, Any] | None = None
    ) -> dict[str, Any]:
        """Make POST request to API endpoint."""
        return await self._auth.request("POST", path, json=json_data)

    async def async_check_version(self) -> dict[str, Any]:
        """Check app version compatibility."""
        return await self._async_get("app/version", {"version": DEFAULT_APP_VERSION})

    async def async_get_user_info(self) -> dict[str, Any]:
        """Get current user information."""
        return await self._async_get("user")

    async def async_get_accounts(self) -> dict[str, Any]:
        """Get list of user accounts."""
        return await self._async_get("accounts")

    async def async_get_account_info(self, account_id: int) -> dict[str, Any]:
        """Get detailed account information by ID."""
        return await self._async_get(f"accounts/{account_id}")

    async def async_get_main_page_debt_info(self) -> dict[str, Any]:
        """Get main page debt information."""
        return await self._async_get("main-page/debt/info")

    async def async_get_information(self, account: str) -> dict[str, Any]:
        """Get general information for an account."""
        return await self._async_get("information", {"account": account})

    async def async_get_counters(self, account: str) -> dict[str, Any]:
        """Get counters (meters) for an account."""
        return await self._async_get("counters", {"account": account})

    async def async_get_balance(self, account: str) -> dict[str, Any]:
        """Get current balance and payment info."""
        return await self._async_get("payments/new-balance", {"account": account})

    async def async_get_counter_readings(
        self, counter_id: str, account: str
    ) -> dict[str, Any]:
        """Get readings history for a specific counter."""
        return await self._async_get(
            f"counters/{counter_id}/readings", {"account": account}
        )

    async def async_send_readings(
        self, account: str, row_id: str, readings: list[str]
    ) -> dict[str, Any]:
        """Send meter readings."""
        if not readings:
            raise RequiredApiParamNotFound("Required API 'readings' parameter not found")
        return await self._async_post(
            "counters/send-readings",
            {
                "account": account,
                "rowId": row_id,
                "readings": readings,
                "platform": DEFAULT_PLATFORM,
            },
        )

    async def async_get_invoice_settings(self, account: str) -> dict[str, Any]:
        """Get invoice email settings for an account."""
        return await self._async_get("invoices/settings", {"account": account})

    async def async_get_invoices(self, account: str, year: int) -> dict[str, Any]:
        """Get list of invoices for an account and year."""
        return await self._async_get(
            "invoices", {"account": account, "year": year}
        )

    async def async_get_invoice_file(self, account: str, date: str) -> dict[str, Any]:
        """Get invoice file as base64 PDF."""
        return await self._async_get(
            "invoices/get-file", {"account": account, "date": date}
        )

    async def async_get_history(
        self, account: str, year: int, month: int
    ) -> dict[str, Any]:
        """Get account history (payments, readings, invoices)."""
        return await self._async_get(
            "history", {"account": account, "year": year, "month": month}
        )



"""Helpers for TNS-Energo API."""
from __future__ import annotations

from base64 import b64encode
from typing import Any

from aiohttp import hdrs

from .const import (
    ACCOUNT_NUMBER_LENGTH,
    API_HASH_HEADER,
    BASE_URL_TEMPLATE,
    BASIC_AUTH_TEMPLATE,
    DEFAULT_API_HASH,
    DEFAULT_CONTENT_TYPE,
    DEFAULT_USER_AGENT,
    DEVICE_ID_HEADER,
)


def is_valid_account(account: str) -> bool:
    """Check if the account number is exactly 12 digits."""
    return len(account) == ACCOUNT_NUMBER_LENGTH and account.isdigit()


def get_base_url(region: str) -> str:
    """Build base URL for the given region."""
    return BASE_URL_TEMPLATE.format(region=region)


def build_request_headers(region: str, device_id: str) -> dict[str, str]:
    """Build common request headers for the given region and device."""
    credentials = BASIC_AUTH_TEMPLATE.format(region=region)
    basic_auth = "Basic " + b64encode(credentials.encode()).decode()
    return {
        hdrs.USER_AGENT: DEFAULT_USER_AGENT,
        hdrs.CONTENT_TYPE: DEFAULT_CONTENT_TYPE,
        API_HASH_HEADER: DEFAULT_API_HASH,
        hdrs.AUTHORIZATION: basic_auth,
        DEVICE_ID_HEADER: device_id,
    }


def is_error_response(response: Any) -> bool:
    """Check if API response indicates an error."""
    if not isinstance(response, dict):
        return True
    return response.get("result") is not True or response.get("statusCode", 200) != 200



"""Exceptions for TNS-Energo API."""
from __future__ import annotations


class TNSEApiError(Exception):
    """Base class for aiotnse errors."""


class TNSEAuthError(TNSEApiError):
    """Base class for aiotnse auth errors."""


class TNSETokenExpiredError(TNSEAuthError):
    """Access token has expired and needs refresh."""


class TNSETokenRefreshError(TNSEAuthError):
    """Refresh token failed, re-authentication required."""


class RegionNotFound(TNSEApiError):
    """Region for account number is not found."""


class RequiredApiParamNotFound(TNSEApiError):
    """Required API parameter not found."""


class InvalidAccountNumber(TNSEApiError):
    """Invalid account number."""



"""TNS-Energo API wrapper."""
from __future__ import annotations

from importlib.metadata import PackageNotFoundError, version

try:
    __version__ = version("aiotnse")
except PackageNotFoundError:
    __version__ = "unknown"

from .api import TNSEApi, async_get_regions
from .auth import AbstractTNSEAuth, SimpleTNSEAuth
from .exceptions import (
    InvalidAccountNumber,
    RegionNotFound,
    RequiredApiParamNotFound,
    TNSEApiError,
    TNSEAuthError,
    TNSETokenExpiredError,
    TNSETokenRefreshError,
)
from .helpers import get_base_url, is_error_response, is_valid_account

__all__ = [
    "AbstractTNSEAuth",
    "InvalidAccountNumber",
    "RegionNotFound",
    "RequiredApiParamNotFound",
    "SimpleTNSEAuth",
    "TNSEApi",
    "TNSEApiError",
    "TNSEAuthError",
    "TNSETokenExpiredError",
    "TNSETokenRefreshError",
    "__version__",
    "async_get_regions",
    "get_base_url",
    "is_error_response",
    "is_valid_account",
]



"""Provide a CLI for TNS-Energo API."""
from __future__ import annotations

import argparse
import asyncio
import logging
import sys
from pprint import pprint
from typing import Any

from aiohttp import ClientError, ClientSession

from . import __version__
from .api import TNSEApi, async_get_regions
from .auth import SimpleTNSEAuth
from .exceptions import TNSEApiError

_LOG_LEVELS = {
    0: logging.ERROR,
    1: logging.WARNING,
    2: logging.INFO,
    3: logging.DEBUG,
}


def _die(message: str) -> None:
    """Print error message and exit."""
    print(f"Error: {message}", file=sys.stderr)
    sys.exit(1)


def _parse_csv(value: str, count: int, flag: str) -> list[str]:
    """Split comma-separated value and validate part count."""
    parts = value.split(",")
    if len(parts) != count:
        _die(f"{flag} expects {count} comma-separated values")
    return parts


def get_arguments() -> argparse.Namespace:
    """Parse command line arguments."""
    parser = argparse.ArgumentParser(description="Command line tool for TNS-Energo API")

    parser.add_argument("--email", help="email for authentication")
    parser.add_argument("--password", help="password for authentication")
    parser.add_argument("--region", help="region code (interactive selection if omitted)")

    parser.add_argument("--accounts", action="store_true", help="list user accounts")
    parser.add_argument("--account-info", type=int, metavar="ID", help="show account details by ID")
    parser.add_argument("--counters", metavar="ACCOUNT", help="show counters for account")
    parser.add_argument("--balance", metavar="ACCOUNT", help="show balance for account")
    parser.add_argument("--readings", metavar="COUNTER_ID,ACCOUNT", help="show counter readings")
    parser.add_argument("--history", metavar="ACCOUNT,YEAR,MONTH", help="show history")
    parser.add_argument("--invoices", metavar="ACCOUNT,YEAR", help="show invoices")
    parser.add_argument("--user", action="store_true", help="show user info")
    parser.add_argument("--regions", action="store_true", help="show available regions")
    parser.add_argument("--version-check", action="store_true", help="check app version")

    sub = parser.add_subparsers(title="send", description="Send the readings command.")
    send = sub.add_parser("send")
    send.add_argument("account", help="account number")
    send.add_argument("row_id", help="row ID from counters response")
    send.add_argument("readings_values", nargs="+", help="reading values (T1, T2, ...)")
    send.set_defaults(command="send")

    parser.add_argument("-v", "--verbose", action="count", default=0, help="increase verbosity level")
    parser.add_argument("-V", "--version", action="version", version=__version__)

    return parser.parse_args()


async def _print_regions(session: ClientSession) -> list[dict[str, Any]]:
    """Fetch and print available regions. Return the region list."""
    data = await async_get_regions(session)
    regions = data["data"]
    print("Available regions:")
    for i, r in enumerate(regions, 1):
        print(f"  {i}. {r['name']} ({r['code']})")
    return regions


async def _select_region(session: ClientSession) -> str:
    """Show region list and let user pick one interactively."""
    regions = await _print_regions(session)
    while True:
        try:
            choice = int(input("\nRegion number: "))
            if 1 <= choice <= len(regions):
                return regions[choice - 1]["code"]
            print(f"Enter a number from 1 to {len(regions)}")
        except ValueError:
            print("Enter a valid number")


async def _send_readings(api: TNSEApi, args: argparse.Namespace) -> None:
    """Confirm and send meter readings."""
    readings = list(args.readings_values)
    print(f"Sending readings for account {args.account}, row {args.row_id}: {readings}")

    if input("Confirm? (y/n): ").lower() != "y":
        print("Cancelled.")
        return

    data = await api.async_send_readings(args.account, args.row_id, readings)
    if data.get("result"):
        print("Readings sent successfully.")
    else:
        print("Failed to send readings.")
        pprint(data)


async def _execute_command(api: TNSEApi, args: argparse.Namespace) -> None:
    """Execute the requested API command."""
    if getattr(args, "command", None) == "send":
        await _send_readings(api, args)
        return

    if args.version_check:
        result = await api.async_check_version()
    elif args.user:
        result = await api.async_get_user_info()
    elif args.accounts:
        result = await api.async_get_accounts()
    elif args.account_info:
        result = await api.async_get_account_info(args.account_info)
    elif args.counters:
        result = await api.async_get_counters(args.counters)
    elif args.balance:
        result = await api.async_get_balance(args.balance)
    elif args.readings:
        counter_id, account = _parse_csv(args.readings, 2, "--readings")
        result = await api.async_get_counter_readings(counter_id, account)
    elif args.history:
        account, year, month = _parse_csv(args.history, 3, "--history")
        result = await api.async_get_history(account, int(year), int(month))
    elif args.invoices:
        account, year = _parse_csv(args.invoices, 2, "--invoices")
        result = await api.async_get_invoices(account, int(year))
    else:
        result = await api.async_get_accounts()

    pprint(result)


async def cli() -> None:
    """Run main."""
    args = get_arguments()
    logging.basicConfig(level=_LOG_LEVELS.get(args.verbose, logging.INFO))

    try:
        async with ClientSession() as session:
            if args.regions:
                await _print_regions(session)
                return

            if not args.email or not args.password:
                _die("--email and --password are required.")

            region = args.region or await _select_region(session)

            auth = SimpleTNSEAuth(
                session, region=region, email=args.email, password=args.password
            )
            print(f"Logging in as {args.email} (region: {region})...")
            await auth.async_login()
            print("Login successful.")

            await _execute_command(TNSEApi(auth), args)
    except ClientError as err:
        _die(f"Connection failed: {err}")
    except TNSEApiError as err:
        _die(f"API error: {err}")



"""TNS-Energo API Auth wrapper."""
from __future__ import annotations

import asyncio
from abc import ABC, abstractmethod
from collections.abc import Callable
from datetime import datetime
from typing import Any

from aiohttp import ClientResponseError, ClientSession

from .const import (
    BEARER_HEADER,
    DEFAULT_API_PATH,
    DEFAULT_PLATFORM,
    DEVICE_ID,
    LOGGER,
)
from .exceptions import TNSEApiError, TNSEAuthError, TNSETokenRefreshError
from .helpers import build_request_headers, get_base_url


class AbstractTNSEAuth(ABC):
    """Abstract class to make authenticated requests."""

    def __init__(self, session: ClientSession, *, region: str) -> None:
        self._session = session
        self._region = region

    @property
    def region(self) -> str:
        """Return current region."""
        return self._region

    @region.setter
    def region(self, value: str) -> None:
        """Set current region."""
        self._region = value

    @property
    def base_url(self) -> str:
        """Return base URL for the current region."""
        return get_base_url(self._region)

    def _build_url(self, path: str) -> str:
        """Build full API URL for the given path."""
        return f"{self.base_url}/{DEFAULT_API_PATH}/{path}"

    def _build_headers(self) -> dict[str, str]:
        """Build common request headers."""
        return build_request_headers(self._region, DEVICE_ID)

    @abstractmethod
    async def async_get_access_token(self) -> str | None:
        """Return a valid access token."""

    async def request(self, method: str, path: str, **kwargs: Any) -> Any:
        """Make a request with proper authorization headers."""
        headers = {**kwargs.pop("headers", {}), **self._build_headers()}

        access_token = await self.async_get_access_token()
        if access_token:
            headers[BEARER_HEADER] = f"Bearer {access_token}"

        url = self._build_url(path)

        LOGGER.debug("Request %s %s", method, url)
        try:
            async with self._session.request(
                method, url, **kwargs, headers=headers, raise_for_status=True
            ) as resp:
                data = await resp.json()
                LOGGER.debug("Response status=%s", resp.status)
        except ClientResponseError as err:
            raise TNSEApiError(
                f"API request failed: {err.status} {err.message}"
            ) from err

        if isinstance(data, dict) and not data.get("result"):
            error = data.get("error", {})
            raise TNSEApiError(error.get("description", "API request failed"))

        return data


class SimpleTNSEAuth(AbstractTNSEAuth):
    """Auth implementation with email/password login and JWT tokens."""

    def __init__(
        self,
        session: ClientSession,
        *,
        region: str,
        email: str | None = None,
        password: str | None = None,
        access_token: str | None = None,
        refresh_token: str | None = None,
        access_token_expires: datetime | None = None,
        refresh_token_expires: datetime | None = None,
        token_update_callback: Callable[[dict[str, Any]], None] | None = None,
    ) -> None:
        """Initialize the auth.

        Two modes:
        1. Login: provide email + password, then call async_login().
        2. Session restore: provide access_token + refresh_token.
        """
        super().__init__(session, region=region)
        self._email = email
        self._password = password
        self._access_token = access_token
        self._refresh_token = refresh_token
        self._access_token_expires = access_token_expires
        self._refresh_token_expires = refresh_token_expires
        self._token_update_callback = token_update_callback
        self._token_lock = asyncio.Lock()

    @property
    def access_token(self) -> str | None:
        """Return current access token."""
        return self._access_token

    @property
    def refresh_token(self) -> str | None:
        """Return current refresh token."""
        return self._refresh_token

    @property
    def access_token_expires(self) -> datetime | None:
        """Return access token expiration time."""
        return self._access_token_expires

    @property
    def refresh_token_expires(self) -> datetime | None:
        """Return refresh token expiration time."""
        return self._refresh_token_expires

    async def async_get_access_token(self) -> str | None:
        """Return a valid access token, refreshing if needed."""
        if self._access_token and (
            not self._access_token_expires
            or datetime.now() < self._access_token_expires
        ):
            return self._access_token

        async with self._token_lock:
            # Re-check after acquiring lock — another coroutine may have refreshed
            if self._access_token and (
                not self._access_token_expires
                or datetime.now() < self._access_token_expires
            ):
                return self._access_token

            # Access token expired or missing — try refresh
            if self._refresh_token and (
                not self._refresh_token_expires
                or datetime.now() < self._refresh_token_expires
            ):
                await self.async_refresh_token()
                return self._access_token

            # Refresh token also expired — try login
            if self._email and self._password:
                await self.async_login()
                return self._access_token

        return self._access_token

    async def _async_auth_request(
        self,
        path: str,
        json_data: dict[str, Any],
        error_class: type[TNSEAuthError],
        default_error: str,
    ) -> dict[str, Any]:
        """Make an auth POST request without Bearer token."""
        url = self._build_url(path)
        LOGGER.debug("Auth request to %s", url)
        try:
            async with self._session.request(
                "POST",
                url,
                json=json_data,
                headers=self._build_headers(),
                raise_for_status=True,
            ) as resp:
                data = await resp.json()
        except ClientResponseError as err:
            raise error_class(
                f"{default_error}: {err.status} {err.message}"
            ) from err

        if not data.get("result"):
            error = data.get("error", {})
            raise error_class(error.get("description", default_error))

        return data

    async def async_login(self) -> dict[str, Any]:
        """Authenticate with email and password."""
        if not self._email or not self._password:
            raise TNSEAuthError("Email and password are required for login")

        data = await self._async_auth_request(
            "user/auth",
            {
                "login": self._email,
                "authType": "email",
                "password": self._password,
                "region": self._region,
                "platform": DEFAULT_PLATFORM,
            },
            TNSEAuthError,
            "Authentication failed",
        )

        token_data = data["data"]
        self._access_token = token_data["accessToken"]
        self._refresh_token = token_data["refreshToken"]
        self._access_token_expires = datetime.fromisoformat(
            token_data["accessTokenExpires"]
        )
        self._refresh_token_expires = datetime.fromisoformat(
            token_data["refreshTokenExpires"]
        )
        self._notify_token_update()

        LOGGER.debug("Login successful")
        return data

    async def async_refresh_token(self) -> dict[str, Any]:
        """Refresh the access token."""
        if not self._refresh_token:
            raise TNSETokenRefreshError("No refresh token available")

        data = await self._async_auth_request(
            "user/refresh-token",
            {"refreshToken": self._refresh_token},
            TNSETokenRefreshError,
            "Token refresh failed",
        )

        self._access_token = data["data"]["accessToken"]
        self._access_token_expires = datetime.fromisoformat(
            data["data"]["accessTokenExpires"]
        )
        self._notify_token_update()

        LOGGER.debug("Token refresh successful")
        return data

    async def async_logout(self) -> dict[str, Any]:
        """Logout and invalidate tokens."""
        data = await self.request("POST", "user/logout")

        self._access_token = None
        self._refresh_token = None
        self._access_token_expires = None
        self._refresh_token_expires = None

        LOGGER.debug("Logout successful")
        return data

    def _notify_token_update(self) -> None:
        """Notify callback about token changes."""
        if self._token_update_callback:
            self._token_update_callback({
                "access_token": self._access_token,
                "refresh_token": self._refresh_token,
                "access_token_expires": (
                    self._access_token_expires.isoformat()
                    if self._access_token_expires
                    else None
                ),
                "refresh_token_expires": (
                    self._refresh_token_expires.isoformat()
                    if self._refresh_token_expires
                    else None
                ),
            })



"""Tests for aiotnse API module."""
from __future__ import annotations

import aiohttp
import pytest
from aioresponses import aioresponses

from aiotnse import TNSEApi
from aiotnse.api import async_get_regions
from aiotnse.const import DEFAULT_APP_VERSION
from aiotnse.exceptions import RequiredApiParamNotFound, TNSEApiError
from tests.common import (
    ACCOUNT,
    ACCOUNT_ID,
    API_URL,
    COUNTER_ID,
    HEADERS,
    ROW_ID,
)
from tests.conftest import load_fixture


class TestAsyncGetRegions:
    async def test_get_regions(self) -> None:
        """Test standalone async_get_regions() without auth."""
        with aioresponses() as mock:
            mock.get(
                f"{API_URL}/contacts/regions",
                payload=load_fixture("regions_response.json"),
                headers=HEADERS,
            )
            async with aiohttp.ClientSession() as session:
                data = await async_get_regions(session)

        assert data["result"] is True
        assert len(data["data"]) > 0
        codes = [r["code"] for r in data["data"]]
        assert "rostov" in codes


    async def test_get_regions_http_error(self) -> None:
        """Test async_get_regions wraps HTTP errors in TNSEApiError."""
        with aioresponses() as mock:
            mock.get(
                f"{API_URL}/contacts/regions",
                status=403,
            )
            async with aiohttp.ClientSession() as session:
                with pytest.raises(TNSEApiError, match="403"):
                    await async_get_regions(session)


class TestTNSEApi:
    async def test_check_version(self, api: TNSEApi, session_mock: aioresponses) -> None:
        session_mock.get(
            f"{API_URL}/app/version?version={DEFAULT_APP_VERSION}",
            payload=load_fixture("app_version_response.json"),
            headers=HEADERS,
        )
        data = await api.async_check_version()

        assert data["result"] is True
        assert data["data"]["status"] == 1

    async def test_get_user_info(self, api: TNSEApi, session_mock: aioresponses) -> None:
        session_mock.get(
            f"{API_URL}/user",
            payload=load_fixture("user_info_response.json"),
            headers=HEADERS,
        )
        data = await api.async_get_user_info()

        assert data["result"] is True
        assert data["data"]["email"] == "user@example.com"
        assert data["data"]["region"] == "rostov"

    async def test_get_accounts(self, api: TNSEApi, session_mock: aioresponses) -> None:
        session_mock.get(
            f"{API_URL}/accounts",
            payload=load_fixture("accounts_response.json"),
            headers=HEADERS,
        )
        data = await api.async_get_accounts()

        assert data["result"] is True
        assert len(data["data"]) == 2
        assert data["data"][0]["id"] == ACCOUNT_ID
        assert data["data"][0]["number"] == ACCOUNT

    async def test_get_account_info(self, api: TNSEApi, session_mock: aioresponses) -> None:
        session_mock.get(
            f"{API_URL}/accounts/{ACCOUNT_ID}",
            payload=load_fixture("account_info_response.json"),
            headers=HEADERS,
        )
        data = await api.async_get_account_info(ACCOUNT_ID)

        assert data["result"] is True
        assert data["data"]["id"] == ACCOUNT_ID
        assert data["data"]["number"] == ACCOUNT
        assert data["data"]["address"]
        assert len(data["data"]["countersInfo"]) > 0

    async def test_get_main_page_debt_info(self, api: TNSEApi, session_mock: aioresponses) -> None:
        session_mock.get(
            f"{API_URL}/main-page/debt/info",
            payload=load_fixture("main_page_debt_response.json"),
            headers=HEADERS,
        )
        data = await api.async_get_main_page_debt_info()

        assert data["result"] is True

    async def test_get_information(self, api: TNSEApi, session_mock: aioresponses) -> None:
        session_mock.get(
            f"{API_URL}/information?account={ACCOUNT}",
            payload=load_fixture("information_response.json"),
            headers=HEADERS,
        )
        data = await api.async_get_information(ACCOUNT)

        assert data["result"] is True

    async def test_get_counters(self, api: TNSEApi, session_mock: aioresponses) -> None:
        session_mock.get(
            f"{API_URL}/counters?account={ACCOUNT}",
            payload=load_fixture("counters_response.json"),
            headers=HEADERS,
        )
        data = await api.async_get_counters(ACCOUNT)

        assert data["result"] is True
        assert len(data["data"]) > 0
        counter = data["data"][0]
        assert counter["counterId"] == COUNTER_ID
        assert counter["rowId"] == ROW_ID
        assert counter["tariff"] == 2
        assert len(counter["lastReadings"]) == 2

    async def test_get_balance(self, api: TNSEApi, session_mock: aioresponses) -> None:
        session_mock.get(
            f"{API_URL}/payments/new-balance?account={ACCOUNT}",
            payload=load_fixture("balance_response.json"),
            headers=HEADERS,
        )
        data = await api.async_get_balance(ACCOUNT)

        assert data["result"] is True
        assert data["data"]["sumToPay"] == 1500.5

    async def test_get_counter_readings(self, api: TNSEApi, session_mock: aioresponses) -> None:
        session_mock.get(
            f"{API_URL}/counters/{COUNTER_ID}/readings?account={ACCOUNT}",
            payload=load_fixture("counter_readings_response.json"),
            headers=HEADERS,
        )
        data = await api.async_get_counter_readings(COUNTER_ID, ACCOUNT)

        assert data["result"] is True
        assert len(data["data"]) > 0
        reading = data["data"][0]
        assert reading["date"]
        assert len(reading["readings"]) == 2

    async def test_send_readings(self, api: TNSEApi, session_mock: aioresponses) -> None:
        session_mock.post(
            f"{API_URL}/counters/send-readings",
            payload=load_fixture("send_readings_response.json"),
            headers=HEADERS,
        )
        data = await api.async_send_readings(ACCOUNT, ROW_ID, ["2690", "1023"])

        assert data["result"] is True

    async def test_send_readings_empty(self, api: TNSEApi) -> None:
        with pytest.raises(RequiredApiParamNotFound):
            await api.async_send_readings(ACCOUNT, ROW_ID, [])

    async def test_get_invoice_settings(self, api: TNSEApi, session_mock: aioresponses) -> None:
        session_mock.get(
            f"{API_URL}/invoices/settings?account={ACCOUNT}",
            payload=load_fixture("invoice_settings_response.json"),
            headers=HEADERS,
        )
        data = await api.async_get_invoice_settings(ACCOUNT)

        assert data["result"] is True
        assert data["data"]["email"] == "user@example.com"
        assert data["data"]["status"] is True

    async def test_get_invoices(self, api: TNSEApi, session_mock: aioresponses) -> None:
        session_mock.get(
            f"{API_URL}/invoices?account={ACCOUNT}&year=2026",
            payload=load_fixture("invoices_response.json"),
            headers=HEADERS,
        )
        data = await api.async_get_invoices(ACCOUNT, 2026)

        assert data["result"] is True
        assert len(data["data"]) > 0
        assert data["data"][0]["date"] == "01.01.2026"

    async def test_get_invoice_file(self, api: TNSEApi, session_mock: aioresponses) -> None:
        session_mock.get(
            f"{API_URL}/invoices/get-file?account={ACCOUNT}&date=01.01.2026",
            payload=load_fixture("invoice_file_response.json"),
            headers=HEADERS,
        )
        data = await api.async_get_invoice_file(ACCOUNT, "01.01.2026")

        assert data["result"] is True
        assert data["data"]["file"]

    async def test_get_history(self, api: TNSEApi, session_mock: aioresponses) -> None:
        session_mock.get(
            f"{API_URL}/history?account={ACCOUNT}&year=2026&month=2",
            payload=load_fixture("history_response.json"),
            headers=HEADERS,
        )
        data = await api.async_get_history(ACCOUNT, 2026, 2)

        assert data["result"] is True
        assert "filters" in data["data"]
        assert "items" in data["data"]
        assert len(data["data"]["items"]) > 0



"""Tests for aiotnse auth module."""
from __future__ import annotations

from base64 import b64encode
from datetime import datetime, timedelta
from unittest.mock import MagicMock

import aiohttp
import pytest
import pytest_asyncio
from aioresponses import aioresponses

from aiotnse import SimpleTNSEAuth
from aiotnse.const import DEVICE_ID, _DEVICE_IDS
from aiotnse.exceptions import TNSEApiError, TNSEAuthError, TNSETokenRefreshError
from tests.common import (
    ACCESS_TOKEN,
    ACCESS_TOKEN_EXPIRES,
    API_URL,
    EMAIL,
    HEADERS,
    PASSWORD,
    REFRESH_TOKEN,
    REFRESH_TOKEN_EXPIRES,
    REGION,
)
from tests.conftest import load_fixture


@pytest_asyncio.fixture
async def login_auth() -> SimpleTNSEAuth:
    """Create a SimpleTNSEAuth instance with login credentials."""
    async with aiohttp.ClientSession() as session:
        yield SimpleTNSEAuth(
            session=session,
            region=REGION,
            email=EMAIL,
            password=PASSWORD,
        )


@pytest_asyncio.fixture
async def no_creds_auth() -> SimpleTNSEAuth:
    """Create a SimpleTNSEAuth instance without credentials."""
    async with aiohttp.ClientSession() as session:
        yield SimpleTNSEAuth(session=session, region=REGION)


class TestSimpleTNSEAuth:
    async def test_login_success(
        self, login_auth: SimpleTNSEAuth, session_mock: aioresponses
    ) -> None:
        session_mock.post(
            f"{API_URL}/user/auth",
            payload=load_fixture("auth_response.json"),
            headers=HEADERS,
        )
        data = await login_auth.async_login()

        assert data["result"] is True
        assert data["statusCode"] == 200
        assert login_auth.access_token == "test_access_token_new"
        assert login_auth.refresh_token == "test_refresh_token_new"
        assert login_auth.access_token_expires == datetime.fromisoformat(
            ACCESS_TOKEN_EXPIRES
        )
        assert login_auth.refresh_token_expires == datetime.fromisoformat(
            REFRESH_TOKEN_EXPIRES
        )

    async def test_login_wrong_credentials(
        self, login_auth: SimpleTNSEAuth, session_mock: aioresponses
    ) -> None:
        session_mock.post(
            f"{API_URL}/user/auth",
            payload=load_fixture("auth_error_response.json"),
            headers=HEADERS,
        )
        with pytest.raises(TNSEAuthError, match="Неверный логин или пароль"):
            await login_auth.async_login()

    async def test_login_no_credentials(self, no_creds_auth: SimpleTNSEAuth) -> None:
        with pytest.raises(TNSEAuthError, match="Email and password are required"):
            await no_creds_auth.async_login()

    async def test_refresh_token_success(
        self, auth: SimpleTNSEAuth, session_mock: aioresponses
    ) -> None:
        session_mock.post(
            f"{API_URL}/user/refresh-token",
            payload=load_fixture("refresh_token_response.json"),
            headers=HEADERS,
        )
        data = await auth.async_refresh_token()

        assert data["result"] is True
        assert auth.access_token == "test_access_token_refreshed"
        assert auth.access_token_expires == datetime.fromisoformat(
            "2026-06-09 21:06:40"
        )

    async def test_refresh_token_no_token(
        self, no_creds_auth: SimpleTNSEAuth
    ) -> None:
        with pytest.raises(TNSETokenRefreshError, match="No refresh token available"):
            await no_creds_auth.async_refresh_token()

    async def test_logout(
        self, auth: SimpleTNSEAuth, session_mock: aioresponses
    ) -> None:
        session_mock.post(
            f"{API_URL}/user/logout",
            payload=load_fixture("logout_response.json"),
            headers=HEADERS,
        )
        data = await auth.async_logout()

        assert data["result"] is True
        assert auth.access_token is None
        assert auth.refresh_token is None
        assert auth.access_token_expires is None
        assert auth.refresh_token_expires is None

    async def test_request_error_response(
        self, auth: SimpleTNSEAuth, session_mock: aioresponses
    ) -> None:
        """Verify request() raises TNSEApiError on error responses."""
        error_payload = {
            "result": False,
            "statusCode": 400,
            "error": {"code": 100, "description": "Some API error"},
        }
        session_mock.get(
            f"{API_URL}/some-endpoint",
            payload=error_payload,
            headers=HEADERS,
        )
        with pytest.raises(TNSEApiError, match="Some API error"):
            await auth.request("GET", "some-endpoint")

    async def test_request_http_error(
        self, auth: SimpleTNSEAuth, session_mock: aioresponses
    ) -> None:
        """Verify request() wraps HTTP errors in TNSEApiError."""
        session_mock.get(
            f"{API_URL}/some-endpoint",
            status=500,
        )
        with pytest.raises(TNSEApiError, match="500"):
            await auth.request("GET", "some-endpoint")

    async def test_login_http_error(
        self, login_auth: SimpleTNSEAuth, session_mock: aioresponses
    ) -> None:
        """Verify async_login wraps HTTP errors in TNSEAuthError."""
        session_mock.post(
            f"{API_URL}/user/auth",
            status=403,
        )
        with pytest.raises(TNSEAuthError, match="403"):
            await login_auth.async_login()

    async def test_basic_auth_rostov(self, auth: SimpleTNSEAuth) -> None:
        expected = "Basic " + b64encode(b"mobile-api-rostov:mobile-api-rostov").decode()
        headers = auth._build_headers()
        assert headers["Authorization"] == expected

    async def test_basic_auth_penza(self) -> None:
        async with aiohttp.ClientSession() as session:
            penza_auth = SimpleTNSEAuth(session=session, region="penza")
        expected = "Basic " + b64encode(b"mobile-api-penza:mobile-api-penza").decode()
        headers = penza_auth._build_headers()
        assert headers["Authorization"] == expected
        assert penza_auth.base_url == "https://mobile-api-penza.tns-e.ru"

    async def test_region_property(self, auth: SimpleTNSEAuth) -> None:
        assert auth.region == REGION

    async def test_set_region(self, auth: SimpleTNSEAuth) -> None:
        auth.region = "nn"
        assert auth.region == "nn"
        assert auth.base_url == "https://mobile-api-nn.tns-e.ru"

    async def test_device_id_in_headers(self, auth: SimpleTNSEAuth) -> None:
        headers = auth._build_headers()
        assert headers["x-device-id"] == DEVICE_ID
        assert DEVICE_ID in _DEVICE_IDS

    async def test_token_update_callback_on_login(
        self, session_mock: aioresponses
    ) -> None:
        """Test that token_update_callback is called after login."""
        callback = MagicMock()
        async with aiohttp.ClientSession() as session:
            auth = SimpleTNSEAuth(
                session=session,
                region=REGION,
                email=EMAIL,
                password=PASSWORD,
                token_update_callback=callback,
            )
            session_mock.post(
                f"{API_URL}/user/auth",
                payload=load_fixture("auth_response.json"),
                headers=HEADERS,
            )
            await auth.async_login()

        callback.assert_called_once()
        token_data = callback.call_args[0][0]
        assert token_data["access_token"] == "test_access_token_new"
        assert token_data["refresh_token"] == "test_refresh_token_new"
        assert token_data["access_token_expires"] is not None
        assert token_data["refresh_token_expires"] is not None

    async def test_token_update_callback_on_refresh(
        self, auth: SimpleTNSEAuth, session_mock: aioresponses
    ) -> None:
        """Test that token_update_callback is called after refresh."""
        callback = MagicMock()
        auth._token_update_callback = callback
        session_mock.post(
            f"{API_URL}/user/refresh-token",
            payload=load_fixture("refresh_token_response.json"),
            headers=HEADERS,
        )
        await auth.async_refresh_token()

        callback.assert_called_once()
        token_data = callback.call_args[0][0]
        assert token_data["access_token"] == "test_access_token_refreshed"
        assert token_data["access_token_expires"] is not None

    async def test_get_access_token_valid(self) -> None:
        """Test async_get_access_token returns token when not expired."""
        async with aiohttp.ClientSession() as session:
            auth = SimpleTNSEAuth(
                session=session,
                region=REGION,
                access_token=ACCESS_TOKEN,
                refresh_token=REFRESH_TOKEN,
                access_token_expires=datetime.now() + timedelta(hours=1),
            )
            token = await auth.async_get_access_token()
        assert token == ACCESS_TOKEN

    async def test_get_access_token_expired_triggers_refresh(
        self, session_mock: aioresponses
    ) -> None:
        """Test async_get_access_token refreshes when access token is expired."""
        async with aiohttp.ClientSession() as session:
            auth = SimpleTNSEAuth(
                session=session,
                region=REGION,
                access_token=ACCESS_TOKEN,
                refresh_token=REFRESH_TOKEN,
                access_token_expires=datetime.now() - timedelta(hours=1),
                refresh_token_expires=datetime.now() + timedelta(days=30),
            )
            session_mock.post(
                f"{API_URL}/user/refresh-token",
                payload=load_fixture("refresh_token_response.json"),
                headers=HEADERS,
            )
            token = await auth.async_get_access_token()
        assert token == "test_access_token_refreshed"

    async def test_get_access_token_both_expired_triggers_login(
        self, session_mock: aioresponses
    ) -> None:
        """Test async_get_access_token re-logs in when both tokens are expired."""
        async with aiohttp.ClientSession() as session:
            auth = SimpleTNSEAuth(
                session=session,
                region=REGION,
                email=EMAIL,
                password=PASSWORD,
                access_token=ACCESS_TOKEN,
                refresh_token=REFRESH_TOKEN,
                access_token_expires=datetime.now() - timedelta(hours=1),
                refresh_token_expires=datetime.now() - timedelta(hours=1),
            )
            session_mock.post(
                f"{API_URL}/user/auth",
                payload=load_fixture("auth_response.json"),
                headers=HEADERS,
            )
            token = await auth.async_get_access_token()
        assert token == "test_access_token_new"

    async def test_get_access_token_no_expiration(self) -> None:
        """Test async_get_access_token returns token when no expiration set."""
        async with aiohttp.ClientSession() as session:
            auth = SimpleTNSEAuth(
                session=session,
                region=REGION,
                access_token=ACCESS_TOKEN,
                refresh_token=REFRESH_TOKEN,
            )
            token = await auth.async_get_access_token()
        assert token == ACCESS_TOKEN



"""Constants for TNS-Energo API."""
from __future__ import annotations

import logging
from random import choice
from typing import Final

LOGGER = logging.getLogger(__package__)

BEARER_HEADER: Final = "authorizationtest"
API_HASH_HEADER: Final = "x-api-hash"
DEVICE_ID_HEADER: Final = "x-device-id"

DEFAULT_CONTENT_TYPE: Final = "application/json"
DEFAULT_USER_AGENT: Final = "Dart/3.9 (dart:io)"
DEFAULT_API_HASH: Final = "b4c9554247f14b9a281f5f60df923f5e"

DEFAULT_API_PATH: Final = "api/v1"
DEFAULT_APP_VERSION: Final = "3.0.12"
DEFAULT_REGION: Final = "rostov"
DEFAULT_PLATFORM: Final = "android"
ACCOUNT_NUMBER_LENGTH: Final = 12

BASIC_AUTH_TEMPLATE: Final = "mobile-api-{region}:mobile-api-{region}"
BASE_URL_TEMPLATE: Final = "https://mobile-api-{region}.tns-e.ru"

_DEVICE_IDS: Final[list[str]] = [
    "AP3A.240905.015",
    "AP3A.240905.015.A2",
    "AP3A.240805.003",
    "AP2A.240705.005",
    "AP2A.240605.004",
    "AP1A.231005.007",
    "UQ1A.240305.002",
    "UQ1A.240105.004",
    "UP1A.231005.007",
    "UP1A.231005.007.A1",
    "TQ3A.230805.001",
    "TQ3A.230705.001",
    "TP1A.221005.002",
    "TP1A.220905.004",
    "SQ3A.220705.003",
    "SP1A.210812.016",
    "SQ1D.220105.007",
    "RQ3A.210805.001",
    "RP1A.201005.004",
    "QQ3A.200805.001",
    "QP1A.190711.020",
    "PQ3A.190801.002",
    "PQ3A.190705.003",
    "PQ2A.190405.003",
    "PQ2A.190305.002",
    "PQ1A.190105.004",
    "PD1A.180720.030",
    "PD1A.180720.031",
]

DEVICE_ID: Final = choice(_DEVICE_IDS)



"""CLI entrypoint for TNS-Energo API."""
from __future__ import annotations

import asyncio

from .cli import cli


def main() -> None:
    """Run the CLI entrypoint."""
    asyncio.run(cli())


if __name__ == "__main__":
    main()
